package farm.ants.agent;
import java.util.*;
import farm.metagent.*;
import utilities.Log;
 
/**
 * An abstract implementation of a particular FSM.
 * The abstract PulseActionMachine handles transition condition
 * monitoring, which involves messages queue, time, shared locks.
 * <p>
 * The abstarct machine also provides access to the PAC, log facilities,
 * and PAM specific state called attributes.
 * <p>
 * Created: Wed Aug  2 14:23:08 2000
 *
 * @author Roger Mailler
 */
public abstract class PulseActionMachine  {

    /**
     * Wait for nothing before transitioning
     */
    public final int WF_NONE = 0;

    /**
     * Wait for messages to arrive
     */
    public final int WF_MESSAGES = 1;

    /**
     * Wait for a specific time
     */
    public final int WF_TIME = 2;

    /**
     * Wait for directory services to return a query
     */
    public final int WF_DIRSERV = 4;

    /**
     * Wait for a negotiation session to terminate
     */
    public final int WF_NEGOTIATION = 8;

    private int stateNum; //holds the current state number of the task
    private String actionType; //a string ID used to differentiate between 
                               //different action types
    private int waitingFor;   //code to determine what we are waiting for
                              //see constants list
    
    private int waitMess = 0;  //used to store the number of messages to wait till
    private int waitTime = 0;  //used to store a time to wait till

    private Hashtable attributes = new Hashtable(); //an object queue for
                                                   //storing state or messages
    private Vector lockList = new Vector();        //list of locks I'm currently waiting for
    private int ID;                                //unique ID number for the action

    private boolean term = false;  //has this machine finished execution

    protected PulseActionController Controller;
    protected Log log = null;


    /**
     * Constructor
     */
    public PulseActionMachine() {
	stateNum = 0;
	attributes.put("Message Queue", new Vector());
	log = Log.getDefault();
    }

    /**
     * Set the PAM type.   This is usually called by the constructor
     * of an implement PAM.
     * @param action The PAM type
     */
    public void setAction(String action){actionType = action;}
    
    /**
     * Returns the PAM type. 
     * @return The PAM type
     */
    public String getAction(){return actionType;}

    /**
     * Sets the unique PAM ID assigned by the controller.   This is usually called by the cotroller
     * when the PAM is registered.
     * @param id The PAM ID
     */
    public void setID(int id){ID = id;}

    /**
     * Returns the PAM ID. 
     * @return The PAM ID
     */
    public int getID(){return ID;}

     /**
     * Sets the state number of the PAM.  This can be used to transition the PAM to a new state
     * @param sNum The new state number
     */    
    public void setState(int sNum){stateNum = sNum;}

    /**
     * Returns the current state number. 
     * @return The state number
     */
    public int getState(){return stateNum;}

    /**
     * Used to tell if the PAM is waiting to transition. 
     * @return True if the waitFor flag is set to anything but WF_NONE
     */
    public boolean isWaiting(){return (waitingFor!=WF_NONE);}

     /**
     * Adds a condition to the current waitFor flag
     * @param condition The condition to add
     */
    public void setWaitFor(int condition){waitingFor |= condition;}

     /**
     * Used to tell if the machine is waiting to transition on a specific condition
     * @param condition The condition to add
     * @return True if the waitFor flag contains that condition
     */
    public boolean isWaitingFor(int condition){return((waitingFor&condition)>0)?true:false;}

     /**
     * Clears the waitFor flag (Sets it to WF_NONE)
     */
    public void clearWaitFor(){waitingFor =WF_NONE;}
 
    /**
     * Returns the waitFor flag
     * @return The value of the waitFor flag
     */
    public int getWaitFor(){return waitingFor;}


    /**
     * Sets the number of messages to wait for
     * @param mNum The number of messages to wait for
     */
    public void setWaitMess(int mNum){waitMess = mNum;}
    
    /**
     * Returns the number of messages that are being waited for
     * @return The number of messages to wait for
     */
    public int getWaitMess(){return waitMess;}

    /**
     * Sets the time to wait for
     * @param time The time to wait for
     */
    public void setWaitTime(int time){waitTime=time;}
    
    /**
     * Returns the waitFor flag
     * @return The vlaue of the waitFor flag
     */
    public int getWaitTime(){return waitTime;}


    /**
     * Called by the controller when the PAm is registered to give a handle to the PAM
     * @param parent The controller that the PAM is registered to
     */
    public void setController(PulseActionController parent){Controller = parent;}


    /******************************************/
    /*      attribute Accessors               */
    /******************************************/

    /**
     * Sets an attribute for the PAM.  Usually this us used to pass
     * data from state to state in the PAM
     * @param attrib The name for the attribute
     * @param value the value to store
     */
    public void setAttribute(String attrib, Object value){attributes.put(attrib, value);}

     /**
     * Returns a previously stored attribute
     * @param attrib The name for the attribute
     * @return the stored attribute object
     */
    public Object getAttribute(String attrib){return attributes.get(attrib);}

     /**
     * Removes a previously stored attribute
     * @param attrib The name for the attribute
     */
    public void removeAttribute(String attrib){attributes.remove(attrib);}

     /**
     * Clears all stored attributes
     */
    public void clearAttributes(){attributes.clear();}

    /**
     * This function causes the PAM to transition to the next state.
     * You specify a transition by passing in the new state number,
     * and the conditions for transitioning.
     * This method does not enfore the WF_LOCK flag, since no locks list is passed in
     * @see #transition(int sNum, int waitf,  int mNum, int time, Vector locks)
     */
    public void transition(int sNum, int waitf, int mNum, int time){
	stateNum = sNum;
	
	waitingFor = waitf;
	waitMess = mNum;
	if(isWaitingFor(WF_MESSAGES) && waitMess == 0){
	    clearWaitFor();
	}

	waitTime = time;
	Vector MQ = (Vector) attributes.get("Message Queue");
	MQ.clear();
    }

    /**
     * This function causes the PAM to transition to the next state.
     * You specify a transition by passing in the new state number,
     * and the conditions for transitioning.
     * @param sNum The state number to transition to
     * @param waitF The waitFor flag setting, use the WF_ constants to set this
     *  flag.  Multiple transition conditions can be specified by OR'ing the individual flags
     *  together.  Transition conditions are always taken as and or relationship.  Therefore, the
     *	first cndition to be satisified will cause the machine to transition.
     * @param mNum The number of messages to wait for before clearing a WF_MESSAGES flag
     * @param time The time to wait for before clearing the WF_TIME flag
     * @param locks The list of locks that must be obtained in order to clear a WF_LOCK flag
     */
    public void transition(int sNum, int waitf, int mNum, 
			   int time, Vector locks){
	stateNum = sNum;	
	waitingFor = waitf;
	waitMess = mNum;
	if(isWaitingFor(WF_MESSAGES) && waitMess == 0){
	    clearWaitFor();
	}

	if(locks.size() > 0){
	    lockList = (Vector) locks.clone();
	    for(int i=0;i<locks.size();i++){
		log.log("Abstract PAM has made the request for the lock: " + ((PACLock) locks.elementAt(i)).toString(), log.LOG_DEBUG2);
		Controller.lock((PACLock) locks.elementAt(i));
	    }
	}
	waitTime = time;
	Vector MQ = (Vector) attributes.get("Message Queue");
	MQ.clear();
    }

    /**
     * This function adds a message to the machines queue.
     * In general if the machine is waiting for messages to arrive
     * this function will store the messages and then when it reaches
     * the total needed tell the machine to execute the next state
     * by clearing the waitFor flag
     * @param message The message to add to the incoming queue
     */
    public void addMessage(Message message){
	Vector messQ = (Vector) attributes.get("Message Queue");

        if (message == null) {
            log.log("Got a null message in PAM addMessage", Log.LOG_ERR);
	}
	else
	    messQ.addElement(message);
	
 	log.log("Recieved a message. Queue size "+ messQ.size(), log.LOG_DEBUG2);
	if (isWaitingFor(WF_MESSAGES) && messQ.size() == waitMess) {
	    clearWaitFor();
        }
    }
    
    /**
     * Called by the controller to grant execution time to the PAM.
     * If the waitFor flag is clear, or the PAM is waiting for on WF_TIME
     * and the time passed in is greater than or equal to the wait for time
     * then executeState is called with the next pending state of the machine.
     * If executeState return true, then the PAM is unregstered from the controller
     * and the terminated flag set to true.
     * @param time The current time (System or simulator)
     */
    public void pulse(int time) {
	//log.log("Got a pulse " + waitTime, log.LOG_INFO);
	if((isWaitingFor(WF_TIME) && time >= waitTime) || (!isWaiting() && lockList.size()==0)){
	    //log.log("Got a pulse " + waitTime, log.LOG_INFO);
	    if(executeState(stateNum)){
		term = true;
		Controller.unregister(this);
	    }
	}	
    }

    /** 
     * This function returns the values of the termianted flag.
     * This flag indicates that at the end of the last pulse, that the PAM
     * completed execution.  
     * @return true if the PAM has been completed
     */   
    public boolean getTerminated(){ 
	return term;
    }

    /** 
     * Used by the controller to notify the PAM that a lock has been obtained.
     * This function removes the lock from the waiting lock list
     * and clear the waitFor flag is the lock list is empty
     * @param lock The lock being granted by the controller
     */   
    public void notifyLock(PACLock lock){
	//the controller has notified me that I have the lock now
	//I also get the id information that was created when I locked
	log.log("PAM " + ID + " has been given the lock for " + lock.toString(), log.LOG_DEBUG);
	lockList.remove(lock);
	log.log("locklist size "+ lockList.size(),log.LOG_DEBUG);
	if(lockList.size() == 0)
	    clearWaitFor();
    }

    /** 
     * Used by the controller to revoke a locking privledge.
     * This function adds the lock back to the waiting lock list
     * and sets the waitFor flag to WF_LOCK
     * @param lock The lock being revoked by the controller
     */    
    public void revokeLock(PACLock lock){
	lockList.add(lock);
    }

    /**
     * Returns a string representation of the PAM
     * @return a string description of the PAM 
     */
    public String toString(){
	String buff =  "ID: " + ID + " Action: " + actionType + " State: " + stateNum + 
	   "\n waitMess: " + waitMess + " waitTime: " + waitTime + "\n";
	//print out the attributes list
	Enumeration attribs = attributes.keys();
	buff = buff + "Attributes: \n";
	if(!attribs.hasMoreElements()){
	    buff = buff + "None\n";
	}
	else{
	    while(attribs.hasMoreElements()){
		buff = buff + attribs.nextElement() + "\n";
	    }
	}
	return buff;
    }

    /**
     * A shorter version of toString()
     * @return a condensed version of toString()
     */
    public String toShortString(){
 	return "ID: " + ID + " Action: " + actionType;
    }

    /**
     * Abstract method that should be overriden to create a PAM.
     * This function is called by the PAM when transitioning form state to state
     * @param sNum The state number to execute
     * @return true if the PAM is done executing, false otherwise
     */
    public abstract boolean executeState(int sNum);



} // PulseActionMachine
