package farm.ants.agent;
import java.util.*;

/**
 * Describes a conflict between a mediating manager and another manager
 * Is used extensively by SPAM to maintain state information in the mediators
 * local search.  This essentially holds the met-level information returned by
 * a TMS reply, hence the parseTMS method.
 * <p>
 * Created: Fri Apr 13 11:05:15 2001
 *
 * @author Roger Mailler
 */
public class TMConflict implements Comparable{
    Vector rNames = new Vector();
    int numAlts = 0;
    int numCons = 0;
    String name;
    ObjFunction objective = new SPAMObjFunction();

    /**
     * Constructor
     * @param n The name of the agent with the conflict
     */
    public TMConflict(String n) {
	name = n;
    }

    /**
     * Adds a resources to the list of conflicts for this agent
     * @param res The resource under conflict
     */
    public void addConflict(String res){
	if(!rNames.contains(res)){
	    //Log.log("Adding new sensor to conflict list:" + sensor,"Server", Log.LOG_ERR);
	    
	    rNames.add(res);
	}
    }

    /**
     * Gets the name of the agent in conflict
     * @return The name
     */
    public String getName(){return name;}

    /** 
     * Gets the number of local conflict for this manager
     * @return The number of conflicts
     */
    public int numLocalConflicts(){return rNames.size();}

    /** 
     * Gets the number of alternative resources for this manager
     * @return The number of resource alternatives
     */
    public int getNumAlts(){return numAlts;}

    /** 
     * Sets the number of alternative resources for this manager
     * @param num The number of alternative resources for this manager
     */    
    public void setNumAlts(int num){numAlts = num;}
    
    /** 
     * Gets a list of the resources under local conflict
     * @return A Vector of resource names
     */ 
    public Vector getConflicts(){return rNames;}

    /** 
     * Sets a list of the resources under local conflict
     * @param cons The list of resource conflicts 
     */    
    public void setConflicts(Vector cons){rNames = cons;}
    
    /** 
     * Gets the number of resource conflicts outside of local consideration
     * @return The number of conflicts
     */    
    public int numAddConflicts(){return numCons;}

    /** 
     * Sets the number of resource conflicts outside of local consideration
     * @param s The number of conflicts
     */  
    public void setAddConflicts(int s){numCons = s;}

    /** 
     * Used to determine is a particular resource is conflicted
     * @param res The resource being checked
     * @return True if it is in conflict
     */
    public boolean hasConflict(String res){return rNames.contains(res);};

    /**
     * Creates a string description of the TMConflict
     * @return The description
     */
    public String toString(){
	return name + " conflicted over: "+ rNames.toString() + " objective: " + objective.getCurrObj().toString() + " num alts: " + numAlts + " num conflicts: " + numCons;  
    }

    /**
     * Return the objective function for this TM
     * @return The function
     */
    public ObjFunction getObjFunc(){
	return objective;
    }

    /**
     * Creates a TMConflict from a properly formated TMS message type
     * @param msg The message to parse
     */
    public void parseTMS(String msg){
	StringTokenizer ST = new StringTokenizer(msg, ",");

	ObjEntry tmpObj = new ObjEntry(ST.nextToken());

	objective.setCurrObj(tmpObj);

	numAlts = Integer.parseInt(ST.nextToken());
	numCons = Integer.parseInt(ST.nextToken());

	while(ST.hasMoreTokens()){
	    this.addConflict(ST.nextToken());
	}
    }

    /**
     * Compares the equality of two TMConflict
     * @param o1 The TMConflcit to compare to
     * @return True if they are the same
     */
    public boolean equals(Object o1){
	return this.toString().equals(((TMConflict) o1).toString());
    }

    /**
     * Compares two TMConflcits based on the number of local conflcits they have
     * @param o1 The TMConflcit to compare to
     * @return -1 if this goes before o1, 0 if equals, 1 otherwise
     */
    public int compareTo(Object o1){
	if(this.numLocalConflicts() < ((TMConflict)o1).numLocalConflicts()){
	    return -1;
	}
	else if(this.numLocalConflicts() > ((TMConflict)o1).numLocalConflicts()){
	    return 1;
	}
	else
	    return 0;
    }

    public Object clone(){
	TMConflict ret = new TMConflict(this.getName());
	ret.getObjFunc().setCurrObj(this.getObjFunc().getCurrObj());
	ret.setNumAlts(this.getNumAlts());
	ret.setAddConflicts(this.numAddConflicts());
	ret.setConflicts((Vector) this.getConflicts().clone());
	return ret;
    }

} // TMConflict
