package farm.bandit.agent;

import farm.simulator.*;
import farm.metagent.*;

import java.util.*;
import utilities.Log;

/**
 * This is the base class used for the learning agents
 * operating in the n-armed bandit domain.  It provides
 * mechanisms for initializing the arm base values,
 * and a function to simulate pulling one of the arms.
 * It also records the statistics for this particular
 * agent, including the reward obtained by the pull,
 * and weather it was optimal or not.
 */
public class BanditAgent extends Agent {
    private static Random baserandom = new Random();
    protected Random random;
    protected int numarms = 10;
    protected double armbasevalues[];
    protected int pulls[];
    protected int totalpulls = 0;
    protected int bestarm = 0;

    /**
     * Constructor, creates the Random object the bandit will use.
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     */
    public BanditAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);

        random = new Random(name.hashCode() + System.currentTimeMillis() + baserandom.nextLong());
    }

    /**
     * Set up the base values for the bandit arms, and
     * keep track of which one is optimal for future use.
     */
    public void begin() {
        super.begin();

        // Make a random
        random = new Random();

        // Generate the base arm values
        armbasevalues = new double[numarms];
        pulls = new int[numarms];
        for (int i = 0; i < numarms; i++) {
            armbasevalues[i] = random.nextGaussian();
            pulls[i] = 0;
            if (armbasevalues[i] > armbasevalues[bestarm]) {
                bestarm = i;
            }
        }
    }

    /**
     * Return the number of arms the bandit has
     */
    public int getNumArms() { return numarms; }

    /**
     * Pulls an arm, returning the reward and updating appropriate state variables
     * This also maintains the aggregate and individual arm pull totals.
     * @param i The arm to pull
     * @return The arm's reward
     */
    public double pullArm(int arm) {

        double reward = armbasevalues[arm] + random.nextGaussian();
	boolean optimal = isOptimalArm(arm);

        pulls[arm]++;
        totalpulls++;

        setProperty(getName() + ":reward", new Float((float)reward));
        setProperty(getName() + ":optimal", new Boolean(optimal));

        //System.err.println(getName() + " Time:" + getTime() + " Arm:" + arm + " Reward:" + reward + " Optimal:" + optimal);

        return reward;
    }

    /**
     * Determines if the arm is optimal or not.  No cheating!
     * @param i The arm to check
     * @return true If the arm is optimal (the best available by base value)
     */
    public boolean isOptimalArm(int arm) {

        return (arm == bestarm);
    }

    /**
     * Returns the number of times an arm has been pulled
     * @param i The arm to check
     */
    public int getNumPulls(int i) { return pulls[i]; }

    /**
     * Returns the total number of times all arms have been pulled
     */
    public int getNumPulls() { return totalpulls; }

    /**
     */
    public void pulse() {
        endRun();
    }
}
