package farm.bandit.agent;

import farm.simulator.*;
import farm.metagent.*;

import java.util.*;
import utilities.Log;

/**
 * This agent implements the pursuit learning algorithm
 * according to (Sutton, Barto) p.43.
 */
public class PursuitAgent extends BanditAgent {
    public static String NAME = "pursuit";
    protected double beta = 0.1;
    protected double q0 = 0;
    protected double q[];
    protected double pi[];

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     */
    public PursuitAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);
    }

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     * @param beta The beta value to use
     */
    public PursuitAgent(FarmMetAgent metagent, String name, double beta) {
        this(metagent, name);

        this.beta = beta;
    }

    /**
     * Initialize the value and probability state vectors
     */
    public void begin() {
        super.begin();

        q = new double[getNumArms()];
        pi = new double[getNumArms()];
        for (int i = 0; i < q.length; i++) {
            q[i] = q0;
            pi[i] = 1d/(double)getNumArms();
        }
    }

    /**
     * Run a single play of the bandit.
     */
    public void pulse() {
        super.pulse();

        // First pick an arm
        float c = random.nextFloat();
        int arm = 0;
        for (int i = 0; i < q.length; i++) {
            arm = i;
            if (c < pi[arm]) {
                break;
            }
            c -= pi[i];
        }

        // Then pull it
        double reward = pullArm(arm);

        // Update q
        q[arm] = q[arm] + (1d/((double)getNumPulls(arm)+1)) * (reward - q[arm]);

        // Find the greedy action a*
        int astar = 0;
        for (int i = 0; i < q.length; i++) {
            if (q[i] > q[astar]) {
                astar = i;
            }
        }

        // Update the probabilities
        for (int i = 0; i < pi.length; i++) {
            if (i == astar)
                pi[i] = pi[i] + beta * (1 - pi[i]);
            else
                pi[i] = pi[i] + beta * (0 - pi[i]);
        }

        //System.err.println(getName() + " Arm:" + arm + " Reward:" + reward + " Optimal:" + isOptimalArm(arm) + " q:" + q[arm] + " pi:" + pi[arm] + " astar:" + astar);
    }
}
