package farm.bandit.plugins;

import java.rmi.*;
import java.rmi.server.*;
import java.io.Serializable;

import utilities.Log;
import farm.simulator.*;
import farm.generic.plugins.*;

import java.util.*;

/**
 * In this component we collect and merge the data
 * generated by the running learning agents.  This is complicated
 * by the fact that a heterogenous collection of agents
 * may be running concurrently, and their data may be
 * distributed across several locations.  This component
 * must then collect the data, separate them out into
 * clusters according to algorithm type, and then compute
 * reward and optimal averages.
 */
public class BanditAnalysis extends FarmPlugIn {

    // Constants
    public static final String PROP_AVERAGEREWARD = "AverageReward";
    public static final String PROP_PERCENTOPTIMAL = "PercentOptimal";

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public BanditAnalysis(String []args)throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Bandit Analysis"));
    }

    /**
     * This is called at the end of a pulse.  Here we create the
     * mobile analysis functions that will be distributed to the various
     * running compoonents.  Locally, they will obtain the relevant
     * data and compute averages.  These averages are then returned
     * to the central analysis component, which computes the
     * final statistics and stored them for public consumption.
     */
    public void pulseEnd() {
        
        try {
            // Compute the average reward statistics
            Map averages = new HashMap();
            LocalFunction fn = new LocalBanditRewardAnalysis();
            for (Iterator i = simulator.getPlugInNames().iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                PlugIn plugin = findPlugIn(name);
                consolidateAverages(averages, (Map)plugin.runLocalFunction(fn));
            }
            consolidateAverages(averages, (Map)simulator.runLocalFunction(fn));
            for (Iterator i = averages.keySet().iterator(); i.hasNext();) {
                String type = (String)i.next();
                Averager avg = (Averager)averages.get(type);
                setProperty(PROP_AVERAGEREWARD + ":" + type, new Float(avg.getAverage()));
            }

            // Compute the percent optimal statistics
            averages = new HashMap();
            fn = new LocalBanditOptimalAnalysis();
            for (Iterator i = simulator.getPlugInNames().iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                PlugIn plugin = findPlugIn(name);
                consolidateAverages(averages, (Map)plugin.runLocalFunction(fn));
            }
            consolidateAverages(averages, (Map)simulator.runLocalFunction(fn));
            for (Iterator i = averages.keySet().iterator(); i.hasNext();) {
                String type = (String)i.next();
                Averager avg = (Averager)averages.get(type);
                setProperty(PROP_PERCENTOPTIMAL + ":" + type, new Float(avg.getAverage()));
            }


        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    public void consolidateAverages(Map local, Map remote) {
        if (remote != null) {
            for (Iterator j = remote.keySet().iterator(); j.hasNext();) {
                String type = (String)j.next();
                Averager ravg = (Averager)remote.get(type);
                
                Averager lavg = (Averager)local.get(type);
                if (lavg == null) {
                    lavg = new Averager();
                    local.put(type, lavg);
                }
                lavg.addAverage(ravg);
            }
        }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new BanditAnalysis(args);
            plugin.run();

        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}

/**
 * Local function which computes the average reward
 * for any available learning algorithms
 */
class LocalBanditRewardAnalysis implements LocalFunction {

    public Object run(PropertyOwner owner) {
        Map averages = new HashMap();

        try {
            // Find average value
            Collection list = owner.findLocalProperties(":reward");
            Iterator i = list.iterator();
            if (!i.hasNext()) return null;
            while (i.hasNext()) {
                String name = (String)i.next();
                String type = new StringTokenizer(name, ":").nextToken();

                Averager avg = (Averager)averages.get(type);
                if (avg == null) {
                    avg = new Averager();
                    averages.put(type, avg);
                }

                Float f = (Float)owner.getLocalProperty(name);
                avg.addValue(f.floatValue());
            }
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }

        return averages;
    }
}

/**
 * Local function which computes the percent optimal
 * for any available learning algorithms
 */
class LocalBanditOptimalAnalysis implements LocalFunction {

    public Object run(PropertyOwner owner) {
        Map averages = new HashMap();
            
        try {
            // Find the percent optimal
            Collection list = owner.findLocalProperties(":optimal");
            Iterator i = list.iterator();
            if (!i.hasNext()) return null;
            while (i.hasNext()) {
                String name = (String)i.next();
                String type = new StringTokenizer(name, ":").nextToken();

                Averager avg = (Averager)averages.get(type);
                if (avg == null) {
                    avg = new Averager();
                    averages.put(type, avg);
                }

                Boolean b = (Boolean)owner.getLocalProperty(name);
                if (b.booleanValue()) avg.addValue(1);
                else avg.addValue(0);
            }
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }

        return averages;
    }
}

/**
 * A simple averaging class which provides a
 * reasonable, portable interface for aggregating the
 * data
 */
class Averager implements Serializable {
    double total = 0;
    int count = 0;

    public void addValue(double value) {
        total += value;
        count++;
    }
    
    public void addAverage(Averager avg) {
        total += avg.total;
        count += avg.count;
    }

    public double getTotal() {
        return total;
    }

    public double getAverage() {
        return (total / (double)count);
    }
}
