package farm.generic.plugins;

import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;

import java.util.*;
import java.io.*;

import utilities.Log;

/**
 * Given a list of numeric properties, computes and stores an incremental
 * average and standard deviation for that property.  Given a property
 * FooBar, it will store the statistics as FooBar:Average and FooBar:StdDev.
 * <p>
 * Parameters:<dl>
 * <dt><tt>-props [comma separated list]</tt>
 * <dd>A list of properties to average
 * <dt><tt>-nostddev</tt>
 * <dd>If true, surpresses the standard deviation calculation
 * </dl>
 */
public class AverageAnalysis extends FarmPlugIn {

    // Constants
    public static final String CLA_PROPERTIES = "props";
    public static final String CLA_NOSTDDEV = "nostddev";

    public static final String AVERAGE_SUFFIX = ":Average";
    public static final String STDDEV_SUFFIX = ":StdDev";

    // Local information
    protected Hashtable averages = new Hashtable();
    protected boolean stddev = true;

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public AverageAnalysis(String []args)throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "AverageAnalysis"));

        stddev = (getCommandLineArgument(args, CLA_NOSTDDEV, null) == null);
        String totrack = getCommandLineArgument(args, CLA_PROPERTIES, null);
        if (totrack != null) {
            for (StringTokenizer tok = new StringTokenizer(totrack, ","); tok.hasMoreTokens(); ) {
                String t = tok.nextToken();
                averageData(t);
            }
        }
    }


    /**
     * Indicates that the specified data name should be averaged.
     * @param name The name of the property key to average
     */
     public void averageData(String name) {
       
        synchronized (averages) {
            log.log("Averaging " + name, Log.LOG_INFO);
            averages.put(name, new Averager());
        }
    }

    /**
     * This is called at the end of a pulse.
     */
    protected void pulseEnd() throws RemoteException {

        // Log the data
        if (timeChanged()) {
            int time = getTime();
            for (Iterator i = averages.keySet().iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Object value = getProperty(name);

                if (value != null && value instanceof Number) {
                    Averager average = (Averager)averages.get(name);
                    average.addValue(((Number)value).doubleValue());
                    setProperty(name + AVERAGE_SUFFIX, new Double(average.getAverage()));
                    if (stddev) {
                        setProperty(name + STDDEV_SUFFIX, new Double(average.getStdDev()));
                    }

                } else {
                    log.log("The property for " + name + "(" + value + ") is not numeric", Log.LOG_ERR);
                }
            }
        }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new AverageAnalysis(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

}

/**
 * A simple averaging class which provides a
 * reasonable, portable interface for aggregating the
 * data
 */
class Averager implements Serializable {
    double total = 0;
    double squares = 0;
    int count = 0;

    public void addValue(double value) {
        total += value;
        squares += value * value;
        count++;
    }
    
    public void addAverage(Averager avg) {
        total += avg.total;
        squares += avg.squares;
        count += avg.count;
    }

    public double getTotal() {
        return total;
    }

    public double getAverage() {
        return (total / (double)count);
    }

    public double getStdDev() {
        if (count > 1) {
            return Math.sqrt(
                             Math.abs(count * squares - total * total) /
                             (count * (count - 1))
                             );
        } else {
            return 0;
        }
    }
}
