package farm.generic.plugins.drivers;

import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;

import java.util.*;

import utilities.Log;

/**
 * This plugin is intended to be used to stop the simulation
 * in a generic way, by providing a simple interface which can
 * be extended to detect the condition under which the simulation
 * should be stopped.
 * <p>
 * This generic class simply looks for the STOPSIMULATION Boolean
 * variable.  If it is present and true, the simulation will be
 * stopped, so other components may halt the run indirectly by
 * setting this flag.  This check is implemented in the
 * simulationShouldStop function, so if a more sophisticated check
 * is required, or if your design calls for the stopping code to
 * be performed in an indpendant component, this funcation may
 * be overriden in a derived class to act in whatever way is
 * appropriate.
 * <p>
 * simulationShouldStop is currently checked from the pulseBegin
 * and pulseEnd functions.
 * <p>
 * @see #simulationShouldStop
 */
public class StopDriver extends FarmPlugIn {

    // Constants
    public static final String STOPSIMULATION = "StopSimulation";

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public StopDriver(String []args) throws RemoteException {
        super(args);
    }

    /**
     * Terminates the simulation
     */
    public void stopSimulation() {

        log.log("Stopping the simulation", Log.LOG_INFO);
        try {
            getSimulator().setStop();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Calls simulationShouldStop to determine if the
     * simulation should be stopped.
     */
    public void pulseBegin() throws RemoteException {

        if (simulationShouldStop()) {
            stopSimulation();
        }
    }

    /**
     * Calls simulationShouldStop to determine if the
     * simulation should be stopped.
     */
    public void pulseEnd() throws RemoteException {

        if (simulationShouldStop()) {
            stopSimulation();
        }
    }

    /**
     * If this function returns true, the simulation will be
     * stopped at the next possible stopping point.
     * <p>
     * Override this function in derived classes to implement
     * your own stop-point detection function.
     * @return true If the simulation should be stopped, false otherwise
     */
    public boolean simulationShouldStop() {

        Boolean stop = (Boolean)getProperty(STOPSIMULATION);
        if (stop != null && stop.booleanValue() == true) {
            return true;
        }

        return false;
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new StopDriver(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
