package farm.sat.agent;

import farm.simulator.*;
import farm.metagent.*;
import farm.sat.plugins.drivers.*;

import java.util.*;
import utilities.Log;

/**
 * This simple agent functions by greedily choosing the best
 * value for it's state at any given time.  During each time window,
 * it will determine what the best value is, by comparing the effects
 * of different values in the context of the constraints and neighboring
 * values it is aware of.  If this best value differs from its current
 * value, it will change its value and notify its neighbors.  No historical
 * state is maintained, so this approach does not necessarily converge
 * on a correct or stable solution.
 */
public class RandomAgent extends Agent {
    VariableModel model;
    Random random = new Random();
    Hashtable neighbors = new Hashtable();

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param m The NodeModel assigned to this agent
     */
    public RandomAgent(FarmMetAgent metagent, VariableModel m) {
	super(metagent, m.getName());

        model = m;
    }

    /**
     */
    public void pulse() {
	super.pulse();

        // What value am i?
        boolean myvalue = model.getValue();

        // Respond to messages
        for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
            checkRun();
	    if (msg.getData().startsWith(">")) {
		Boolean b = Boolean.valueOf(msg.getData().substring(1));
                neighbors.put(msg.getSource(), b);
            }
        }

        // Pick a new color
        boolean bestvalue = chooseBestValue();
        if (myvalue != bestvalue && random.nextInt(4) == 0) {
            log.log(getAgentName() + " changing value from " + myvalue + " to " + bestvalue, Log.LOG_INFO);
            model.setValue(bestvalue);
            for (Iterator i = model.getRelated(); i.hasNext(); ) {
                sendMessage(new Message(getAgentName(), (String)i.next(), ">" + bestvalue));
                checkRun();
            }
        }

        endRun();
    }

    /**
     * Selects the best color, greedily based on local knowledge.
     * This function will return the value which produces the fewest
     * conflicts with known constraints.
     */
    public boolean chooseBestValue() {
        int t = 0;
        int f = 0;

        // Make the value list
        Map map = new HashMap();
        for (Enumeration e = neighbors.keys(); e.hasMoreElements(); ) {
            String name = (String)e.nextElement();
            map.put(name, neighbors.get(name));
        }

        // Check it
        map.put(model.getName(), Boolean.TRUE);
        for (Iterator i = model.getClauses(); i.hasNext(); ) {
            Clause c = (Clause)i.next();
            if (c.isSatisfiedBy(map)) {
                t++;
            }
        }
        map.put(model.getName(), Boolean.FALSE);
        for (Iterator i = model.getClauses(); i.hasNext(); ) {
            Clause c = (Clause)i.next();
            if (c.isSatisfiedBy(map)) {
                f++;
            }
        }

        // What's best?
        if (t == f) {
            if (t == model.numClauses())
                return model.getValue();
            else
                return random.nextBoolean();
        } else {
            return (t > f ? true : false);
        }
    }
}
