package farm.sat.plugins.drivers; 

import java.rmi.*;
import java.rmi.server.*;

import farm.simulator.*;

import java.util.*;
import java.io.*;
import java.awt.geom.*;
import utilities.Log;
import farm.sat.plugins.*;

/**
 * <p>
 * Properties:<dl>
 * <dt><tt>-variables [integer]</tt>
 * <dd>The number of variables to use (default 10)
 * <dt><tt>-clauses [integer]</tt>
 * <dd>The number of clauses to generate (default 15)
 * <dt><tt>-length [integer]</tt>
 * <dd>The length of the generated clauses (default 3)
 * <dt><tt>-datafile [string]</tt>
 * <dd>The name of the configuration data file to use (if any)
 * <dt><tt>-type [awc,random]</tt>
 * <dd>The type of agent to produce (default random)
 * <dt><tt>-metagents [integer]</tt>
 * <dd>The integer number of metagents to create, used for intial load 
 *     balancing (default 1)
 * </dl>
 */
public class SATDriver extends FarmPlugIn {

    // Constants
    /**
     * Command line identifier for the number of variables
     */
    public static final String CLA_VARIABLES = "variables";

    /**
     * Command line identifier for the number of clauses
     */
    public static final String CLA_CLAUSES = "clauses";

    /**
     * Command line identifier for the length of the clauses
     */
    public static final String CLA_LENGTH = "length";

    /**
     * Command line identifier for the configuration source file
     */
    public static final String CLA_DATAFILE = "datafile";

    /**
     * Command line identifier for the type of agent to create
     */
    public static final String CLA_TYPE = "type";

    /**
     * Command line identifier for number of metagents in the environmemt
     */
    public static final String CLA_METAGENTS = "metagents";

    /**
     * Property identifier for the clause list
     */
    public static final String PROP_CLAUSES = "Clauses";

    /**
     * Property identifier for the clause list
     */
    public static final String PROP_VARIABLES = "Variables";

    /**
     * Property identifier to indicate driver if finshed with initial startup
     */
    public static final String CLA_DRIVERSTARTUP = "dstart";

    int numvars = 10;
    int numclauses = 15;
    int clauselength = 3;
    ArrayList clauses = new ArrayList();
    int metaAgents = 1;
    String type = "random";
    String datafile = null;

    /**
     * Constructor
     * @param args An array of command line arguments for the driver
     */
    public SATDriver(String []args) throws RemoteException {
	super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "SAT Driver"));
        setVariables(Integer.valueOf(getCommandLineArgument(args, CLA_VARIABLES, String.valueOf(numvars))));
        setClauses(Integer.valueOf(getCommandLineArgument(args, CLA_CLAUSES, String.valueOf(numclauses))));
        setLength(Integer.valueOf(getCommandLineArgument(args, CLA_LENGTH, String.valueOf(clauselength))));
        setType(getCommandLineArgument(args, CLA_TYPE, type));
        setDataFile(getCommandLineArgument(args, CLA_DATAFILE, datafile));
        setMetaAgents(Integer.valueOf(getCommandLineArgument(args, CLA_METAGENTS, String.valueOf(metaAgents))));
    }

    /**
     * Sets the number of variables
     * @param i Number of variables
     */
    public void setVariables(Integer i) {
        numvars = i.intValue();
        setProperty(CLA_VARIABLES, i);
    }

    /**
     * Sets the number of clauses
     * @param i Number of clauses
     */
    public void setClauses(Integer i) {
        numclauses = i.intValue();
        setProperty(CLA_CLAUSES, i);
    }

    /**
     * Sets the clause length
     * @param i The clause length
     */
    public void setLength(Integer i) {
        clauselength = i.intValue();
        setProperty(CLA_LENGTH, i);
    }

    /**
     * Sets the type of agent to produce
     * @param t The agent type (e.g. random, awc)
     */
    public void setType(String t) {
        type = t;
        setProperty(CLA_TYPE, t);
    }

    /**
     * Sets the data file source for SAT configuration
     * @param f The name of the file
     */
    public void setDataFile(String f) {
        datafile = f;
        if (datafile != null) {
            setProperty(CLA_DATAFILE, f);
        }
    }

    /**
     * Sets the number of metagents
     * @param i The number of metagents
     */
    public void setMetaAgents(Integer i) {
        metaAgents = i.intValue(); 
        setProperty(CLA_METAGENTS, i);
    }    

    /**
     * Starts the plugin, creates the varables, creates the clauses,
     * assigns the variables to metagents
     */
    public void begin() throws RemoteException {
	super.begin();

        // Generate the nodes
        // Set the clauses
        Hashtable clauselists = new Hashtable();
        ArrayList variables = new ArrayList(numvars);
        for (int i = 0; i < numvars; i++) {
            variables.add(new VariableModel(this, "ND"+i));
            clauselists.put("ND"+ i, new ArrayList());
        }

        // Generate the clauses and set their data
        if (datafile != null) {
            clauses = readClauses(datafile);
        } else {
            clauses = createClauses(numvars, numclauses, clauselength);
        }

        // Set the clauses
        for (Iterator i = clauses.iterator(); i.hasNext();) {
            Clause c = (Clause)i.next();
            log.log("Clause: " + c, Log.LOG_INFO);
            for (Iterator j = c.getTerms(); j.hasNext(); ) {
                Term t = (Term) j.next();
		ArrayList al = (ArrayList) clauselists.get(t.variable);
		al.add(c);
            }
        }
        for (int i = 0; i < numvars; i++) {
            VariableModel model = (VariableModel)variables.get(i);
            model.setClauses((ArrayList)clauselists.get(model.getName()));
        }
        setProperty(PROP_CLAUSES, clauses);

        // Assign the targets
        log.log("Creating assignments for " + metaAgents + " metagents", Log.LOG_INFO);
        int div = (int)Math.floor(numvars / metaAgents);
        int mod = numvars % metaAgents;
        for (int i = metaAgents-1, start = 0; i >= 0; i--) {
            String metkey = "MA" + i  + farm.metagent.FarmMetAgent.AGENTS_DESCRIPTION;
            int finish = start + div-1;
            if (mod > 0) {
                finish++;
                mod--;
            }
            if (finish >= start) {
                String assign = type + " " + start + "-" + finish;
                String prev = (String)getProperty(metkey);
                if (prev != null)
                    assign = prev + " " + assign;
                setProperty(metkey, assign);
            }
            start = finish + 1;
        }
       setProperty(CLA_DRIVERSTARTUP, new Boolean(true));

    }

    /**
     * Creates clauses
     */
    public ArrayList createClauses(int vars, int clauses, int length) {
        ArrayList clist = new ArrayList(clauses);
        boolean vlist[] = new boolean[clauses];

        Random random = getRandom();

        // To ensure satisfiability, first generate the variables
        for (int i = 0; i < vars; i++) {
            vlist[i] = random.nextBoolean();
        }

        // Then the clauses
        log.log("Creating " + numclauses + " clauses of length " + length + " with " + vars + " variables", Log.LOG_INFO);
        for (int i = 0; i < clauses; i++) {
            Clause c = new Clause();
            int valid = random.nextInt(length);
            for (int l = 0; l < length; l++) {
                int t;
                do {
                    t = random.nextInt(vars);
                } while (c.containsVariable("ND" + t));
                if (l == valid) {
                    c.addTerm(vlist[t], "ND" + t);
                } else {
                    c.addTerm(random.nextBoolean(), "ND" + t);
                }
            }
            clist.add(c);
        }

        return clist;
    }

    /**
     * Reads clauses in from a file
     */
    public ArrayList readClauses(String filename) {
        ArrayList clauses = new ArrayList();

        try {
            log.log("Reading clause definitions from " + filename, Log.LOG_INFO);
            FileReader reader = new FileReader(filename);
            BufferedReader buf = new BufferedReader(reader);
            for (String str = buf.readLine(); (str = buf.readLine()) != null;) {
                str = str.trim();
                if (str.startsWith("c")) continue;
                if (str.startsWith("p")) continue;
                if (str.startsWith("%")) break;
                str = str.substring(0, str.length() - 2);
                Clause c = new Clause(str);
                clauses.add(c);
            }

        } catch (IOException e) {
        }

        return clauses;
    }

    /**
     * Main Function
     */
    public static void main(String[] args) {
	try {
	    FarmPlugIn plugin = new SATDriver(args);
	    plugin.run();
	} catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
