package farm.sat.plugins.drivers; 

import farm.simulator.*;

import java.util.*;
import java.io.*;

/**
 * A model for describing a graph node.  Has accessors
 * for the node's name, neighbors and color
 */
public class VariableModel implements Serializable {
    protected FarmPlugIn plugin;
    protected String name;
    protected ArrayList clauses = new ArrayList();
    protected boolean value = false;
    protected int priority = 0;
    private boolean undefined = true;

    /**
     * Constructor
     * @param p Some plugin
     * @param n The node name
     * @param b Initial value
     */
    public VariableModel(FarmPlugIn p, String n, boolean b) {
        plugin = p;

        name = n;

        setValue(b);
    }

    /**
     * Constructor
     * @param p Some plugin
     * @param n The node name
     */
    public VariableModel(FarmPlugIn p, String n) {
        this(p, n, false);
    }

    /**
     * Returns the node's name
     */
    public String getName() { return name; }

    /**
     * Returns the variable's value
     */
    public boolean getValue() { return value; }

    /**
     * Sets the node's color
     */
    public void setValue(boolean b) { 

        if (value != b || undefined) {
            value = b;
            plugin.setProperty(name + ":value", value ? Boolean.TRUE : Boolean.FALSE); 
            undefined = false;
        }
    }

    /**
     * Returns the variable's priority
     */
    public int getPriority() { return priority; }

    /**
     * Sets the variable's priority
     */
    public void setPriority(int p) { 

        if (priority != p) {
            priority = p;
            plugin.setProperty(name + ":priority", new Integer(priority)); 
        }
    }

    /**
     * Returns the clauses this variable is involved in
     */
    public Iterator getClauses() {
        clauses = (ArrayList)plugin.getProperty(name + ":clauses");
        return clauses.iterator();
    }

    /**
     * Returns the number of clauses this variable is involved in
     */
    public int numClauses() {
        clauses = (ArrayList)plugin.getProperty(name + ":clauses");
	return clauses.size();
    }

    /**
     * Adds a clause related to this variable
     */
    public void addClause(Clause c) {

        if (c != null) {
            clauses.add(c);
        }
        plugin.setProperty(name + ":clauses", clauses);        
    }

    /**
     * Sets the agent's clauses to those in the collection.
     */
    public void setClauses(Collection c) {

        clauses.clear();
        clauses.addAll(c);
        plugin.setProperty(name + ":clauses", clauses);        
    }

    /**
     * Returns an iterator of related variables (i.e. those that exist in
     * clauses with this one.  Returned elements are String representations
     * of the variable numbers.
     */
    public Iterator getRelated() {
        ArrayList list = new ArrayList();
	//getClauses();
        for (Iterator i = clauses.iterator(); i.hasNext(); ) {
            Clause c = (Clause)i.next();
            for (Iterator j = c.getTerms(); j.hasNext(); ) {
                Term t = (Term)j.next();
                String n = t.variable;
                if (!list.contains(n)) {
                    list.add(n);
                }
            }
        }

        return list.iterator();
    }

    /**
     * Returns the list of clauses that this variable
     * shares with another
     * @param v the other variable
     * @return The list of shared clauses
     */
    public ArrayList howRelated(VariableModel v) {
	return howRelated(v.getName());
    }

    /**
     * Returns the list of clauses that this variable
     * shares with another
     * @param v the other variable
     * @return The list of shared clauses
     */
    public ArrayList howRelated(String v) {
	ArrayList list = new ArrayList();
	//getClauses();
	for (Iterator i = clauses.iterator(); i.hasNext(); ) {
            Clause c = (Clause)i.next();
            if(c.containsVariable(v)){
		list.add(c);
	    }
	}
	return list;
    }

    /**
     * Stringify
     */
    public String toString() {
        return getName() + " (" + getValue() + ")"; 
    }
}
