/**
 */

package farm.metagent;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;

import farm.simulator.*;
import utilities.Log;
import utilities.Clock;

/**
 * The FarmMetAgent class acts as the basic metagent, from which other,
 * more specialized metagents can be derived.  It is a "special" plugin,
 * which is treated differently by the simulator in that it will provide
 * a PULSE_IN pulse during which agents are expected to operate.  The 
 * FarmMetAgent itself provides the necessary mechanisms to do this,
 * along with means of delivering and accepting messages, caching
 * local property information and managing agent lists.
 * <p>
 * The metagents primary purpose is to act as a container for Agents,
 * which are the real actors in the system.  On startup (in the begin)
 * function, createAllAgents is called, which should instantiate the
 * initial set of agents to be managed by the metagent.  During
 * execution the metagent behaves just like any other plugin, with the
 * additional deliver of the PULSE_IN pulse type, which is used to
 * allocate deliberate amounts of execution time to the Agents 
 * under its control.
 * <p>
 * Properties:<dl>
 * <dt><tt>-msgmintime [integer]</tt>
 * <dd>The minimum time a message will take to reach its destination (default 0, requires msgmaxtime)
 * <dt><tt>-msgmaxtime [integer]</tt>
 * <dd>The maximum time a message will take to reach its destination (default 0)
 * <dt><tt>-msgbytetime [integer]</tt>
 * <dd>The amount of time per byte of content data added to message send duration (default 0)
 * <dt><tt>-msglossrate [float]</tt>
 * <dd>The probability (0-1) that a message is lost in transit (default 0)
 * <dt><tt>-endrun [boolean]</tt>
 * <dd>If false, endRun calls will not be allowed by agents (default true)
 * <dt><tt>-realtime [boolean]</tt>
 * <dd>If true, time is interpreted as ms to run, if false time is considered
 * to be the number of checkRun calls that can be made (default true)
 * </dl>
 * @see Agent
 */
public class FarmMetAgent extends FarmPlugIn implements MetAgent {

    // Constants
    public static final String AGENTS_DESCRIPTION = "_AgentsDescription";
    public static final String CLA_MSGMINTIME = "msgmintime";
    public static final String CLA_MSGMAXTIME = "msgmaxtime";
    public static final String CLA_MSGBYTETIME = "msgbytetime";
    public static final String CLA_MSGLOSSRATE = "msglossrate";
    public static final String CLA_ENDRUN = "endrun";
    public static final String CLA_REALTIME = "realtime";

    // Agents
    private Hashtable agents = new Hashtable();
    private Vector metagents = new Vector();

    // For messages
    private Hashtable routes = new Hashtable();
    private Vector incoming = new Vector();

    // Utilities
    protected Random random = new Random();

    // Local information
    private Hashtable cachedproperties = new Hashtable();
    private Vector toupdate = new Vector();
    private int msgmintime = 0;
    private int msgmaxtime = 0;
    private int msgbytetime = 0;
    private float msglossrate = 0f;
    private boolean allowendrun = true;
    private boolean realtime = true;
    private double timescale = 1.0;

    /**
     * Constructor
     * @param simulator The remote simulator object to use
     * @param name The name of the metagent
     */
    public FarmMetAgent(String []args) throws RemoteException {
        super(args);

        // Init
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Unknown MetAgent"));
        setMessageMinTime(Integer.parseInt(getCommandLineArgument(args, CLA_MSGMINTIME, String.valueOf(getMessageMinTime()))));
        setMessageMaxTime(Integer.parseInt(getCommandLineArgument(args, CLA_MSGMAXTIME, String.valueOf(getMessageMaxTime()))));
        setMessageByteTime(Integer.parseInt(getCommandLineArgument(args, CLA_MSGBYTETIME, String.valueOf(getMessageByteTime()))));
        setMessageLossRate(Float.parseFloat(getCommandLineArgument(args, CLA_MSGLOSSRATE, String.valueOf(getMessageLossRate()))));
        setAllowEndRun(Boolean.valueOf(getCommandLineArgument(args, CLA_ENDRUN, String.valueOf(getAllowEndRun()))).booleanValue());
        setRealTime(Boolean.valueOf(getCommandLineArgument(args, CLA_REALTIME, String.valueOf(getRealTime()))).booleanValue());
    }

    /**
     * Accessors
     */
    public String getMetAgentName() { return getPlugInName(); }

    public void setMessageMinTime(int i) { msgmintime = i; }
    public int getMessageMinTime() { return msgmintime; }
    public void setMessageMaxTime(int i) { msgmaxtime = i; }
    public int getMessageMaxTime() { return msgmaxtime; }
    public void setMessageByteTime(int i) { msgbytetime = i; }
    public int getMessageByteTime() { return msgbytetime; }
    public void setMessageLossRate(float f) { msglossrate = f; }
    public float getMessageLossRate() { return msglossrate; }
    public void setAllowEndRun(boolean b) { allowendrun = b; }
    public boolean getAllowEndRun() { return allowendrun; }
    public void setRealTime(boolean b) { realtime = b; }
    public boolean getRealTime() { return realtime; }

    public void setBaselineCPUMetric(double gmetric) {
        Linpack bench = new Linpack();
        double lmetric = bench.doBenchmark();
        timescale = gmetric / lmetric;
        log.log("Local benchmark was " + lmetric + " (global is " + gmetric + "), new time scale value is " + timescale, Log.LOG_INFO);
    }

    /**
     * Registers a remote metagent with this one.  This is used as part
     * of the message delivery system.  When a message is to be delivered,
     * the routing table is first searched for an appropriate destination.
     * If none is found, the known list of remote metagents is searched.
     * When an appropriate destination is found, that connection is
     * recorded in the routing table.
     * @param agent The metagent to register
     */
    public void registerMetAgent(MetAgent agent) {

        synchronized (metagents) {
            try {
                log.log("Adding metagent " + agent.getPlugInName(), Log.LOG_INFO);
                metagents.addElement(agent);
            } catch (RemoteException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        }
    }

    /**
     * Unregisters a remote metagent
     * @param agent The metagent to unregister
     */
    public void unregisterMetAgent(MetAgent agent) {

        synchronized (metagents) {
            try {
                log.log("Removing metagent " + agent.getPlugInName(), Log.LOG_INFO);
                metagents.removeElement(agent);
            } catch (RemoteException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        }
    }

    /**
     * Looks through the outgoing messages for each agent
     * and routes them to their correct destination, according
     * to their scheduled delivery time.
     */
    public void routeMessages() {
        int total = 0;
        int time = -1;
	int byteTotal = 0;

        synchronized (agents) {
            for (Enumeration e = agents.elements(); e.hasMoreElements();) {
                Agent agent = (Agent)e.nextElement();
                for (Message msg = agent.peekFirstOutgoingMessage(); msg != null; msg = agent.peekFirstOutgoingMessage()) {

                    // See if it is corrupted
                    if (msg.isCorrupted()) {
                        log.log("Dropped sending message: " + msg, Log.LOG_INFO);
                        msg = agent.getFirstOutgoingMessage();
                        continue;
                    }

                    // Ensure it is ready to be sent
                    if (time < 0) time = getTime();
                    if (msg.getActualSendTime() > time) break;

                    // Send it
                    msg = agent.getFirstOutgoingMessage();
                    total++;
		    byteTotal += msg.getData().length();
                    routeMessage(msg);
                }                
            }
        }
        
        // Record the total
        setProperty(getMetAgentName() + ":MessagesSent", new Integer(total));
        setProperty(getMetAgentName() + ":BytesSent", new Integer(byteTotal));
    }
    
    /**
     * Delivers a message to the appropriate metagent, who
     * is then responsible for delivering it to the appropriate
     * local agent.
     * @param msg The message to deliver
     */
    public void routeMessage(Message msg) {
        MetAgent metagent = (MetAgent)routes.get(msg.getDestination());
        
        // No route?  Find one.
        if (metagent == null) {
            // Is it mine?
            if (hasAgent(msg.getDestination())) {
                metagent = this;

            // Is it someone else's?
            } else for (Iterator i = metagents.iterator(); i.hasNext(); ) {
                metagent = (MetAgent)i.next();
                try {
                    if (metagent.hasAgent(msg.getDestination())) {
                        break;
                    }
                } catch (RemoteException e) {
                    System.err.println("Exception: " + e.getMessage());
                    e.printStackTrace();
                }
                metagent = null;
            }
            
            // Save this for later
            if (metagent != null) {
                routes.put(msg.getDestination(), metagent);
            }
        }

        // Deliver it if we can
        if (metagent != null) {
            try {
                metagent.acceptMessage(msg);
            } catch (RemoteException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        } else {
            log.log("Warning: Could not find suitable metagent to route message to: " + msg, Log.LOG_WARNING);
        }
    }

    /**
     * Accepts a message from a remove message delivery
     * service, and queues it for local delivery to the
     * intended target.
     * @param msg The message to receive
     */
    public void acceptMessage(Message msg) {

        synchronized (incoming) {
	    //log.log("incoming message: " + msg, Log.LOG_INFO);
	    incoming.addElement(msg);
        }
    }

    /**
     * Delivers messages stored in the incoming queue to their
     * intended destinations.  Messages are only delivered if
     * their actual receive time is equal or less than
     * the current time.
     */
    public void deliverMessages() {
        int total = 0;
	int byteTotal = 0;
        synchronized (incoming) {
            if (!incoming.isEmpty()) {
                int time = getTime();

                for (Iterator i = incoming.iterator(); i.hasNext(); ) {
                    Message msg = (Message)i.next();

                    // See if it is corrupted
                    if (msg.isCorrupted()) {
                        log.log("Dropped received message: " + msg, Log.LOG_INFO);
                        i.remove();
                        continue;
                    }

                    // Ensure it is time for delivery
                    if (msg.getActualReceiveTime() > time) continue;

                    // Find the agent
                    Agent a = (Agent)getAgent(msg.getDestination());
                    //log.log("delivering message: " + msg, Log.LOG_INFO);
                    if (a != null) {
                        a.receiveMessage(msg);
		        byteTotal += msg.getData().length();
                        total++;
                    } else {
                        log.log("Warning: Could not find destination for message: " + msg, Log.LOG_WARNING);
                    }
                    i.remove();
                }
            }
        }

        // Record the total
        setProperty(getMetAgentName() + ":MessagesReceived", new Integer(total));
        setProperty(getMetAgentName() + ":BytesReceived", new Integer(byteTotal));
    }

    /**
     * Calculates and sets the send and delivery times, and any other
     * pertinant characteristics, of a message being sent by an agent.
     * <p>
     * This function takes on part of the communication modeling role,
     * by assigning actual send and receive times for the message.  This will be
     * the send time of the most recently queued message (if any),
     * plus a random value from the minimum to maximum message time
     * (if any), plus a per-byte penalty (if any).  It also determines
     * the probability of corruption.
     * <p>
     * Domains which require alternate communication characteristics should
     * override this function, and potentially the routing and delivery
     * functions, to provide them.
     *
     * @param source The source of the message
     * @param msg The message being sent
     */
    public void prepareMessageDelivery(Agent source, Message msg) {
        int time = getTime();

        // Determine if the message is lost or not
        if (random.nextFloat() < getMessageLossRate()) {
            msg.setCorrupted(true);
        }

        // Calculate the send time, based on a strict sequential
        // ordering of messages.
        Message lastmsg = source.peekLastOutgoingMessage();
        if (lastmsg != null) {
            time = lastmsg.getActualReceiveTime();
        }
        msg.setActualSendTime(time);

        // Calculate the receive time
        int range = getMessageMaxTime() - getMessageMinTime();
        int delay = getMessageMinTime() + ((range > 0) ? random.nextInt(range) : 0);
        if (getMessageByteTime() > 0) {
            delay += getMessageByteTime() * msg.getData().length();
        }
        msg.setActualReceiveTime(time + delay);
    }

    /**
     * Handles a property request.  This will use the cached value if
     * it exists, or fetch a remote value if needed
     * @param name The name of the property to access
     */
    public synchronized Object getProperty(String name) {
        Object value = null;

        /*if (cachedproperties.containsKey(name)) {
            value = cachedproperties.get(name);
            } else {*/
            value = super.getProperty(name);
            /* if (value != null) {
                cachedproperties.put(name, value);
            }
            }*/

        return value;
    }

    /**
     * Handles a property set command.  This will update the local
     * cached value and set it back to the simulator.
     * @param name The name of the property to access
     */
    public boolean setProperty(String name, Object value) {

        /*cachedproperties.put(name, value);
        synchronized (toupdate) {
            if (!toupdate.contains(name)) {
                toupdate.addElement(name);
            }
            }*/

        return super.setProperty(name, value);
    }

    /**
     * Updates the owner copy of local cached properties, and resets the
     * local update vector
     */
    public void updateProperties() {

        synchronized (toupdate) {
            for (Iterator i = toupdate.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                super.setProperty(name, cachedproperties.get(name));
                i.remove();
            }
        }
        synchronized (cachedproperties) {
            cachedproperties.clear();
        }
    }

    /**
     * Adds a new agent to the metagent
     * @param a The agent to add
     */
    public void addAgent(Agent a) {

        synchronized (agents) {
            agents.put(a.getAgentName(), a);
            a.start();
            log.log("Added agent " + a.getAgentName(), Log.LOG_INFO);
        }
    }

    /**
     * Gets a particular agent
     * @param name The name of the agent to get
     * @return The agent, or null if not found
     */
    public Agent getAgent(String name) {

        return (Agent)agents.get(name);
    }

    /**
     * Determines if the metagent owns the particular
     * agent.
     * @param name The name of the agent in question
     * @return True if the metagent owns the agent, false otherwise
     */
    public boolean hasAgent(String name) {

        return agents.containsKey(name);
    }

    /**
     * Gets a list of the available agents' names
     * @return An enumeration of agent names
     */
    public Enumeration getAllAgentNames() {
        Vector v = new Vector();

        synchronized (agents) {
            for (Iterator i = agents.keySet().iterator(); i.hasNext(); ) {
                v.addElement(i.next());
            }
        }

        return v.elements();
    }

    /**
     * Override this with your function that should create
     * the agents to be controlled by this metagent.  This
     * function will be called once during the begin phase,
     * and may be called later if a reset occurs.
     * @see #addAgent
     */
    public void createAllAgents() {

        // First flush existing agents
        stopAllAgents();

        // Then create the new ones
        int numagents = 10;
        log.log("Creating " + numagents + " agents", Log.LOG_INFO);
        for (int i = 0; i < numagents; i++) {
            addAgent(new Agent(this, "A" + i));
        }
    }

    /**
     * Stops and removes all agents from the metagent
     */
    public void stopAllAgents() {

        // Tell the agents to stop
        for (Enumeration e = agents.elements(); e.hasMoreElements();) {
            Agent agent = (Agent)e.nextElement();
            agent.setStop();
        }

        // Wait for them to complete
        for (Iterator i = agents.entrySet().iterator(); i.hasNext();) {
            Agent agent = (Agent)((Map.Entry)i.next()).getValue();
            while (agent.isAlive()) {
            	agent.setRun(0, realtime, null);
                mysleep(50);
            }
            i.remove();
        }
   }

    /**
     * This registeres the metagent with the farm simulator, and
     * creates the initial set of Agent objects.
     * #see createAllAgents
     */
    public void begin() throws RemoteException {
        
        // Parent
        super.begin();

        // Register
        simulator.registerMetAgent(this);

        // Create the agents
	long stime = System.currentTimeMillis();
        createAllAgents();
        log.log("Created agents in " + (System.currentTimeMillis() - stime) + " ms", Log.LOG_INFO);
    }

    /**
     * Call this to perform a "pulse" of time
     */
    protected void pulseIn(int time) {
        int pulsetime = time;

        // Are we done?
        if (isStopping() || (time <= 0)) return;

        // Scale the time
        if (realtime) {
            pulsetime = (int)Math.round(time * timescale);
            log.log("Pulsing for " + pulsetime + " ms (scaled from " + time + ")");
        } else {
            pulsetime = time;
            log.log("Pulsing for " + pulsetime + " clicks");
        }

	long stime = System.currentTimeMillis();

        // Pulse the agents
        int count = 0;
        try {
            synchronized (agents) {
                count = pulseAgentsSequential(pulsetime);
                //count = pulseAgentsParallel(pulsetime);
            }
        } catch (Exception e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }

	log.log("Done pulsing " + count + " agents, time: " + (System.currentTimeMillis() - stime));
    }

    /**
     * Pulses the agents in series.  Individual agents are pulsed in random
     * order each time this is called.
     */
    public int pulseAgentsSequential(long time) {
        Vector localagents = new Vector(agents.values());

        // Start them up
        int count = 0;
        while (!localagents.isEmpty()) {
            Boolean notify = new Boolean(false);
            Agent agent = (Agent)localagents.elementAt(random.nextInt(localagents.size()));
            if (!agent.shouldStop()) {
                if (agent.isAlive()) {
                    synchronized (notify) {
                        agent.setRun(time, realtime, notify);
                        try {
                            notify.wait();
                        } catch (InterruptedException e) {
                            throw new IllegalStateException("Interrupted");
                            }
                    }
                    /*while (agent.isRunning()) {
                        mysleep((int)(time / 4));
                        }*/
                    count++;
                }
            }
            localagents.removeElement(agent);
        }

        return count;
    }

    /**
     * Pulses the agents in parallel
     */
    public int pulseAgentsParallel(long time) {

        // Start them up
        int count = 0;
        for (Enumeration e = agents.elements(); e.hasMoreElements();) {
            Agent agent = (Agent)e.nextElement();
            agent.setRun(time, realtime, null);
            count++;
        }

        // Wait for the agents to finish
        while (true) {
            int active = 0;
            for (Enumeration e = agents.elements(); e.hasMoreElements();) {
                Agent agent = (Agent)e.nextElement();
                if (agent.isRunning()) {
                    active++;
                }
            }
            if (active == 0) break;
            log.log("Waiting for " + active + " agents", Log.LOG_INFO);
            mysleep((int)(time / 4 * active));
        }

        return count;
    }

    /**
     * This is called at the beginning of a pulse, before the
     * agents are started.
     */
    protected void pulseBegin() {

        // Set the time
        Integer i = (Integer)getProperty(Simulator.P_TIME);
        if (i != null) {
            Clock.setTime(i.intValue());
        }

        // Deliver messages
        deliverMessages();
    }

    /**
     * This is called at the end of a pulse, after the agents
     * should have completed.
     */
    protected void pulseEnd() {

        // Route messages
        routeMessages();

        // Update properties
        updateProperties();
    }

    /**
     * Called when the metagent is to end
     */
    public void end() throws RemoteException {

        // Unregister
        simulator.unregisterMetAgent(this);

        // Stop the agents
        stopAllAgents();

        // Stop the parent
        super.end();
   }

    /**
     * Convenience function
     */
    public void mysleep(int i) {

        try {
            Thread.currentThread().yield();
            Thread.currentThread().sleep(i);
        } catch (InterruptedException e) { }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmMetAgent agent = new FarmMetAgent(args);
            agent.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
