/**
 */

package farm.simulator;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;

import farm.simulator.*;
import utilities.Log;

/**
 * This is the basic plugin class.  All other plugins that should connect to
 * the farm simulation will usually be derived from this class.  It provides
 * the basic farm plugin functionality, including connecting to the simulator,
 * registering itself, providing a control flow shell (begin, end and pulse* 
 * methods), and simple accessors (getProperty and getTime).
 * <p>
 * Note that the parameters listed below are available to all derived plugins
 * as well (this is generally true of parameters for all plugins and their
 * relatives).
 * <p>
 * Properties:<dl>
 * <dt><tt>-name [string]</tt>
 * <dd>The name by which the plugin should be known
 * <dt><tt>-host [host address]</tt>
 * <dd>The name or ip address of the machine hosting the simulator
 * <dt><tt>-port [integer]</tt>
 * <dd>The port number the simulator is listening at
 * <dt><tt>-seed [integer]</tt>
 * <dd>The random number seed to use.  If zero, no defined seed is used (default 0)
 * </dl>
 */
public class FarmPlugIn extends UnicastRemoteObject implements PlugIn {

    // Constants
    /**
     * Property identifier for the plugin's name
     */
    public static final String CLA_NAME = "name";

    /**
     * Property identifier for the random number seed
     */
    public static final String CLA_SEED = "seed";

    // Utilities
    protected Simulator simulator = null;
    protected Log log = null;
    
    // Pulse timing
    protected Boolean pulselock = new Boolean(true);
    protected Boolean pulsedone = new Boolean(true);

    // Local information
    private Thread shutdownHook;
    private Hashtable properties = new Hashtable();
    private Hashtable propertytokens = new Hashtable();
    private Hashtable plugins = new Hashtable();

    private Hashtable Qk = new Hashtable();
    private Hashtable rk = new Hashtable();
    private float alpha = 0.3f;

    protected int pulse = Simulator.PULSE_NONE;
    private boolean done = false;
    private boolean dead = false;
    private volatile int runtime = 0;
    private volatile int lasttime = -1;
    private String name = "Unknown PlugIn";
    private int seed = 0;
    private Random random;

    private DataToken cachetoken;

    private boolean logprops = false;
    private boolean greedyreaders = false;
    private boolean firstreader = false;
    private boolean greedywriters = false;
    private boolean firstwriter = false;
    private boolean aggressivepush = false;
    private boolean rollingpush = false;

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public FarmPlugIn(String args[]) throws RemoteException {
        super();

        // Get log
        log = Log.getDefault();

        // Add the shutdown hook
        try {
            Runtime.getRuntime().addShutdownHook(shutdownHook = new Thread("ShutdownHook") {
                    public void run() {
                        setStop();
                        while (!dead) {
                            mysleep(100);
                        }
                    }
                });
        } catch (NoSuchMethodError ex) {
            log.log("Failed to addShutdownHook, you may be running an older version of java", Log.LOG_WARNING);
        }

        // Set the security manager
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        // Connect to the simulator
        for (int i = 0; i < FarmSimulator.SIMULATOR_PORTS.length; i++) {
            int port = FarmSimulator.SIMULATOR_PORTS[i];
            try {
                String host = getCommandLineArgument(args, Simulator.CLA_SIMULATORHOST, "127.0.0.1");
                if (i == 0)
                    port = Integer.parseInt(getCommandLineArgument(args, Simulator.CLA_SIMULATORPORT, String.valueOf(port)));
                String name = "//" + host + ":" + port + "/simulator";
                simulator = (Simulator)Naming.lookup(name);
                break;
                
            } catch (Exception e) {
                System.err.println("Error binding to port " + port + ((i+1 < FarmSimulator.SIMULATOR_PORTS.length) ? ", trying port " + FarmSimulator.SIMULATOR_PORTS[i+1] : ", no more ports to try"));
            }
        }

        // Init
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Unknown Plugin"));
        setSeed(Integer.valueOf(getCommandLineArgument(args, CLA_SEED, String.valueOf(seed))));

        // Property stuff
        if (getCommandLineArgument(args, "gw", null) != null) greedywriters = true;
        if (getCommandLineArgument(args, "gr", null) != null) greedyreaders = true;
        if (getCommandLineArgument(args, "ap", null) != null) aggressivepush = true;
        if (getCommandLineArgument(args, "fr", null) != null) firstreader = true;
        if (getCommandLineArgument(args, "fw", null) != null) firstwriter = true;
        if (getCommandLineArgument(args, "rp", null) != null) rollingpush = true;
        log.log( "GW:" + greedywriters +
                " GR:" + greedyreaders +
                " AP:" + aggressivepush +
                " RP:" + rollingpush +
                " FW:" + firstwriter +
                " FR:" + firstreader);
    }

    /**
     * Returns the plugin's name
     */
    public String getPlugInName() { return name; }

    /**
     * Sets the plugin's name
     */
    public void setPlugInName(String n) { 
        name = n;
        if (log != null) {
            log.setName(name);
        }
    }

    /**
     * Sets the random number seed, and creates the random object
     * @param i The seed
     */
    public void setSeed(Integer i) {
        seed = i.intValue();

        if (seed == 0) {
            random = new Random();
        } else {
            random = new Random(seed);
        }
    }

    /**
     * Returns the random number generator
     * @return The Random object
     */
    public Random getRandom() {
        return random;
    }

    /**
     * Returns the plugin's reference to the simulator
     */
    public Simulator getSimulator() { return simulator; }

    /**
     * Parses command line arguments.
     * @see farm.simulator.FarmSimulator#getCommandLineArgument
     */
    public String getCommandLineArgument(String []args, String name, String def) {

        return FarmSimulator.getCommandLineArgument(args, name, def);
    }

    /**
     * Gets a handle to a plugin
     */
    public PlugIn findPlugIn(String name) {
        PlugIn plugin = null;

        // Cached?
        if (plugins.containsKey(name)) {
if (logprops) System.err.println("LOCAL FIND PLUGIN\t" + name);
            plugin = (PlugIn)plugins.get(name);

        // Not seen before?
        } else {
            if (name.equals(getPlugInName())) {
if (logprops) System.err.println("LOCAL FIND PLUGIN\t" + name);
                plugin = this;
            } else {
                try {
                    plugin = simulator.findPlugIn(name);
if (logprops) System.err.println("REMOTE FIND PLUGIN\t" + name);
                } catch (RemoteException e) {
                    System.err.println("Exception: " + e.getMessage());
                    e.printStackTrace();
                }
            }
            if (plugin != null) {
                plugins.put(name, plugin);
            }
        }

        return plugin;
    }

    /**
     * Handles a remote property request.
     * @param name The name of the property to access
     */
    public Object getPropertyRemote(String name) {
        Object value = null;
        DataToken token = null;

        // If it has a token, check it
        if ((token = (DataToken)propertytokens.get(name)) != null
            && token.isOwner()) {
            
            value = getLocalProperty(name);
        } else {
            log.log("Hey, I don't own " + name);
            value = new DataToken(Simulator.NAME);
        }

        return value;
    }

    /**
     * Gets the local property value
     * @param name The name of the property to access
     */
    public Object getLocalProperty(String name) {

        synchronized (properties) {
            return properties.get(name);
        }
    }

    /**
     * Handles a property request.  This will use the cached value if
     * it exists, or fetch a remote value if needed
     * @param name The name of the property to access
     * @return The property's value, or null if not found
     */
    public Object getProperty(String name) {
        Object value = null;
        boolean locallyread = false;
        DataToken token = null;
        String pushrequest = null;

if (logprops) System.err.println("READING\t" + name + "\t" + getPlugInName());
        long stime = System.currentTimeMillis();

        // Save access
        if (rollingpush) {
            Integer accesses = (Integer)rk.get(name);
            if (accesses == null) {
                accesses = new Integer(1);
                synchronized (Qk) {
                    if (!Qk.containsKey(name)) Qk.put(name, new Float(0));
                }
            } else {
                accesses = new Integer(accesses.intValue() + 1);
            }
            rk.put(name, accesses);
        }

        // Greedy readers
        if (greedyreaders) {
            if ((token = (DataToken)propertytokens.get(name)) == null ||
                (!token.isTransient() && !token.isOwner())) {
                try {
                    simulator.setPropertyOwner(name, getPlugInName(), null);
                } catch (RemoteException e) {
                    e.printStackTrace();
                }
            }
        }

        // If it has a token, check it
        for (int i = 0; i < 5; i++) {
            token = null;
            pushrequest = null;

            try {
                if ((token = (DataToken)propertytokens.get(name)) != null) {
                    // If I'm the owner, I should have it
                    if (token.isOwner()) {
                        if (logprops) System.err.println("DATA LOCAL GET\t" + name + "\t" + getPlugInName());
                        value = getLocalProperty(name);
                        locallyread = true;
                        // If I have a cached value, I might just use it
                    } else if (token.isTransient()) {
                        if (logprops) System.err.println("DATA LOCAL CACHE\t" + name + "\t" + getPlugInName());
                        value = getLocalProperty(name);
                        locallyread = true;
                        // If someone else is the owner, they should have it
                    } else if (token.getSource() != null) {
                        // Find the source
                        PlugIn plugin = token.getSourcePlugIn();
                        if (plugin == null) {
                            plugin = findPlugIn(token.getSource());
                            token.setSourcePlugIn(plugin);
                        }
                        // The plugin has it
                        if (plugin != null) {
if (logprops) System.err.println("DATA REMOTE GET\t" + name + "\t" + getPlugInName() + "\t" + token.getSource() + "\t" + System.currentTimeMillis());
                            if (logprops) System.err.println("DATA REMOTE GET\t" + name + "\t" + getPlugInName() + "\t" + token.getSource());
                            value = plugin.getPropertyRemote(name);
                            pushrequest = token.getSource();
                            // The Simulator has it
                        } else if (token.getSource().equals(Simulator.NAME)) {
                            value = simulator.getPropertyRemote(name);
                            if (logprops) System.err.println("DATA REMOTE GET\t" + name + "\t" + getPlugInName() + "\t" + Simulator.NAME);
                            // First reader
                            if (firstreader && !(value instanceof DataToken)) {
                                try {
                                    simulator.setPropertyOwner(name, getPlugInName(), null);
                                } catch (RemoteException e) {
                                    e.printStackTrace();
                                }
                            }
                            // No one has it (dangit)
                        } else {
                            log.log("Could not find source plugin " + token.getSource() + " for property " + name);
                            value = new DataToken(Simulator.NAME);
                        }
                    }

                    // If no token, ask simulator
                } else {
                    value = simulator.getPropertyRemote(name);
                    if (logprops) System.err.println("DATA REMOTE GET\t" + name + "\t" + getPlugInName() + "\t" + Simulator.NAME);
                    // First reader
                    if (firstreader && !(value instanceof DataToken)) {
                        try {
                            simulator.setPropertyOwner(name, getPlugInName(), null);
                        } catch (RemoteException e) {
                            e.printStackTrace();
                        }
                    }
                }
            } catch (RemoteException e) {
                System.err.println("Exception: " + e.getMessage());
                value = new DataToken(Simulator.NAME);
            }

            // Are we being redirected?
            if (value != null && value instanceof DataToken) {
                token = (DataToken)value;
                propertytokens.put(name, token);
            } else {
                break;
            }
        }

        // Aggressive pushing
        if (aggressivepush && pushrequest != null) {
            PlugIn plugin = findPlugIn(pushrequest);
            try {
                long sptime = System.currentTimeMillis();
                plugin.startPushProperty(name, getPlugInName());
if (logprops) System.err.println("DATA OVERHEAD PUSH START\t" + name + "\t" + pushrequest + "\t" + (System.currentTimeMillis() - sptime));
            } catch (RemoteException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        }

        // Sanity check
        if (value instanceof DataToken) {
            log.log("Could not find source for property " + name);
            value = null;
        }

        if (logprops) System.err.println("READ\t" + name + "\t" + getPlugInName() + "\t" + (System.currentTimeMillis() - stime));

        return value;
    }

    /**
     * Handles a property set command made by a remote plugin,
     * this is typically not called locally.
     * @param name The name of the property to access
     * @param value The data to store there
     * @return null if the property was set, a DataToken redirect otherwise
     */
    public DataToken setPropertyRemote(String name, Object value) {
        boolean set = false;
        DataToken token = null;

        // If it has a token, check it
        if ((token = (DataToken)propertytokens.get(name)) != null
            && token.isOwner()) {
            
            setLocalProperty(name, value);
            set = true;
        }

        if (set)
            return null;
        else
            return new DataToken(Simulator.NAME);
   }

    /**
     * Sets a local property.
     * @param name The name of the property to access
     * @param value The data to store there
     */
    public void setLocalProperty(String name, Object value) {
        DataToken token = null;

        synchronized (properties) {
            if (value != null) {
                properties.put(name, value);
            } else {
                properties.remove(name);
            }
if (logprops) System.err.println("DATA LOCAL SET\t" + name + "\t" + System.currentTimeMillis());
            if ((token = (DataToken)propertytokens.get(name)) != null
                && token.isOwner()) {
                pushProperty(name, value, token);
            }
       }
    }

    /**
     * Handles a property set command.  This will update the local
     * cached value.
     * @param name The name of the property to access
     */
    public boolean setProperty(String name, Object value) {
        boolean set = false;
        DataToken token = null;

if (logprops) System.err.println("WRITING\t" + name + "\t" + getPlugInName());
        long stime = System.currentTimeMillis();

        // Greedy writers
        if (greedywriters) {
            if ((token = (DataToken)propertytokens.get(name)) == null ||
                !token.isOwner()) {
                try {
                    simulator.setPropertyOwner(name, getPlugInName(), null);
                } catch (RemoteException e) {
                    e.printStackTrace();
                }
            }
        }

        for (int i = 0; i < 5; i++) {
            try {
                // If it has a token, check it
                token = null;
                if ((token = (DataToken)propertytokens.get(name)) != null) {
                    // If I'm the owner, I should have it
                    if (token.isOwner()) {
                        setLocalProperty(name, value);
                        set = true;
                        if (logprops) System.err.println("DATA LOCAL SET\t" + name + "\t" + getPlugInName());
                        // If someone else is the owner, they should be notified
                    } else if (token.getSource() != null) {
                        // Find the source
                        PlugIn plugin = token.getSourcePlugIn();
                        if (plugin == null) {
                            plugin = findPlugIn(token.getSource());
                            token.setSourcePlugIn(plugin);
                        }
                        // Someone else has it
                        if (plugin != null) {
                            if (logprops) System.err.println("DATA REMOTE SET\t" + name + "\t" + getPlugInName() + "\t" + token.getSource());
                            if ((token = plugin.setPropertyRemote(name, value)) == null) {
                                set = true;
                            }
                        // The simulator has it
                        } else if (token.getSource().equals(Simulator.NAME)) {
                            if ((token = simulator.setPropertyRemote(name, value)) == null) {
                                set = true;
                                // First writer
                                if (firstwriter) {
                                    try {
                                        simulator.setPropertyOwner(name, getPlugInName(), null);
                                    } catch (RemoteException e) { e.printStackTrace(); }
                                }
                            }
                            if (logprops) System.err.println("DATA REMOTE SET\t" + name + "\t" + getPlugInName() + "\t" + Simulator.NAME);
                        // No one has it (dangit)
                        } else {
                            log.log("Could not find source plugin " + token.getSource() + " for property " + name);
                            token = new DataToken(Simulator.NAME);
                        }
                    } else {
                        log.log("Could not find owner plugin to set property property " + name);
                    }
                } else {
                    if ((token = simulator.setPropertyRemote(name, value)) == null) {
                        set = true;
                        // First writer
                        if (firstwriter) {
                            try {
                                simulator.setPropertyOwner(name, getPlugInName(), null);
                            } catch (RemoteException e) { e.printStackTrace(); }
                        }
                    }
                    if (logprops) System.err.println("DATA REMOTE SET\t" + name + "\t" + getPlugInName() + "\t" + Simulator.NAME);
                }
            } catch (RemoteException e) {
                System.err.println("Error setting property " + name + ": " + value);
                e.printStackTrace();
                token = new DataToken(Simulator.NAME);                
            }

            if (!set) {
                if (token == null)
                    propertytokens.remove(name);
                else
                    propertytokens.put(name, token);
            } else {
                break;
            }

        }

        if (!set) {
            log.log("Error. was unable to write value for " + name + " (token " + token + ")");
        }

        if (logprops) System.err.println("WRITE\t" + name + "\t" + getPlugInName() + "\t" + (System.currentTimeMillis() - stime));

	return set;
    }

    /**
     * Can be used to determine what properties exist that
     * match a particular form
     * @param name The name of the property to search for
     */
    public Collection findProperties(String name) {
        
        try {
            return simulator.findProperties(name);

        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
        
        return java.util.Collections.EMPTY_LIST;
    }

    /**
     * Finds local property names which match the provided name.
     * @param name The name of the property to search for
     */
    public Collection findLocalProperties(String name) {
        ArrayList matches = new ArrayList();
		
        for (Iterator i = properties.keySet().iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            if (s.indexOf(name) >= 0) {
                matches.add(s);
            }
        }

        return matches;
    }
    
    /**
     * Called when a remote entity wants this plug in to own 
     * a particular property
     * @param name The property name
     * @param value The most up-to-date value for the property
     * @param pushees The (optional) list of targets to which this property should be pushed
     */
    public void ownProperty(String name, Object value, Collection pushees) {

        synchronized (properties) {
            // Make the token
            DataToken token = new DataToken();
            if (pushees != null) {
                for (Iterator i = pushees.iterator(); i.hasNext(); ) {
                    token.addCacheTarget((String)i.next());
                }
            }
            propertytokens.put(name, token);
if (logprops) System.err.println("DATA OVERHEAD OWN\t" + name + "\t" + getPlugInName());

            // Store the value
            setLocalProperty(name, value);
        }
    }

    /**
     * Called when a remote entity wants this plug in to stop owning
     * a particular property
     * @param name The property name
     * @return The value which was stored by this plugin
     */
    public Object disownProperty(String name) {
        Object value = null;

        synchronized (properties) {
            DataToken token = (DataToken)propertytokens.remove(name);
            synchronized (token) {
                if (token != null) {
                    for (Iterator i = token.getCacheTargets(); i.hasNext(); ) {
                        PlugIn plugin = findPlugIn((String)i.next());
                        if (plugin != null) {
                            try {
                                plugin.uncacheProperty(name);
                            } catch (RemoteException e) {
                                log.log("Error uncaching " + name + " at plugin");
                            }
                        }
                    }
                }
            }
            value = properties.remove(name);
if (logprops) System.err.println("DATA OVERHEAD DISOWN\t" + name + "\t" + getPlugInName());
        }

        return value;
    }

    /**
     * Retrieves the plugin's data token for a particular property
     * @param name The property name
     * @return The data token, or null if none found
     */
    public DataToken getPropertyToken(String name) {

        return (DataToken)propertytokens.get(name);
    }

    /**
     * Called when a remote entity wants to give this plugin a local
     * cached version of a particular property
     * @param name The property name
     * @param value The cached value for the property
     * @param token The property's data token
     */
    public void cacheProperty(String name, Object value, DataToken token) {

        propertytokens.put(name, token);
        setLocalProperty(name, value);

        // Save access
        if (rollingpush) {
            Integer accesses = (Integer)rk.get(name);
            if (accesses == null) {
                accesses = new Integer(-1);
                synchronized (Qk) {
                    if (!Qk.containsKey(name)) Qk.put(name, new Float(0));
                }
            } else {
                accesses = new Integer(accesses.intValue() - 1);
            }
            rk.put(name, accesses);
        }
    }

    /**
     * Called when a remote entity wants to tell this plugin they
     * should no longer cache a particular property
     * @param name The property name
     */
    public void uncacheProperty(String name) {

        propertytokens.remove(name);
        setLocalProperty(name, null);
    }

    /**
     * Cleans up the property list by removing transients
     */
    public void cleanProperties() {

        synchronized (properties) {

            for (Iterator i = propertytokens.keySet().iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                DataToken token = (DataToken)propertytokens.get(name);
                /*if (token.isTransient()) {
                    properties.remove(name);
                    i.remove();
                    }*/
            }
        }

        if (rollingpush) {
            long stime = System.currentTimeMillis();
            synchronized (Qk) {
                for (Iterator i = Qk.keySet().iterator(); i.hasNext(); ) {
                    String name = (String)i.next();
                    float q = ((Float)Qk.get(name)).floatValue();
                    int r = 0;

                    if (rk.containsKey(name))
                        r = ((Integer)rk.get(name)).intValue();
                
                    q = q + alpha * (r - q);

                    DataToken token = (DataToken)propertytokens.get(name);
                    int push = 0;
                    if (token != null &&
                        !token.isOwner()) {
                        if (q > 0.3 &&
                            !token.isTransient()) {
                            push = 1;
                        } else if (q < 0 &&
                                   token.isTransient()) {
                            push = -1;
                        }
                    }

                    //if (logprops) System.err.println("ROLLING\t" + name + "\t" + q + "\t" + push + "\t" + r);
                    if (push != 0) {
                        // Find the source
                        PlugIn plugin = token.getSourcePlugIn();
                        if (plugin == null) {
                            plugin = findPlugIn(token.getSource());
                            token.setSourcePlugIn(plugin);
                        }
                    
                        // Push it
                        try {
                            long sptime = System.currentTimeMillis();
                            if (push > 0) {
                                plugin.startPushProperty(name, getPlugInName());
                                if (logprops) System.err.println("DATA OVERHEAD PUSH START\t" + name + "\t" + token.getSource() + "\t" + (System.currentTimeMillis() - sptime));
                            } else if (push < 0) {
                                plugin.stopPushProperty(name, getPlugInName());
                                if (logprops) System.err.println("DATA OVERHEAD PUSH STOP\t" + name + "\t" + token.getSource() + "\t" + (System.currentTimeMillis() - sptime));
                            }
                        } catch (RemoteException e) {
                            System.err.println("Exception: " + e.getMessage());
                            e.printStackTrace();
                        }
                    }

                    rk.put(name, new Integer(0));
                    Qk.put(name, new Float(q));
                }
            }
if (logprops) System.err.println("ROLLING CHECK\t" + (System.currentTimeMillis() - stime));
        }
    }

    /**
     * Adds a plugin to a property's cache target list
     * @param name The property name
     * @param pluginname The plugin's name
     * @return True if the push request was satisfied
     */
    public boolean startPushProperty(String name, String pluginname) {
        DataToken token = null;
        boolean added = false;

        if ((token = (DataToken)propertytokens.get(name)) != null &&
            token.isOwner()) {
            PlugIn plugin = findPlugIn(pluginname);
            try {
                if (plugin != null) {
                    long sptime = System.currentTimeMillis();
                    if (cachetoken == null) cachetoken = new DataToken(true, getPlugInName());
                    plugin.cacheProperty(name, getLocalProperty(name), cachetoken);
                    token.addCacheTarget(pluginname);
                    added = true;
                    if (logprops) System.err.println("DATA OVERHEAD REMOTE PUSH\t" + name + "\t" + pluginname + "\t" + (System.currentTimeMillis() - sptime));
               } else {
                    log.log("Could not find plugin " + pluginname + " to push " + name + " to");
                }
            } catch (RemoteException e) {
                log.log("Error pushing " + name + " to pluginname");
            }
        } else {
            log.log("Request push start for " + name + " from " + pluginname + " denied because I am not the owner");
        }

        return added;
    }

    /**
     * Removes a plugin from a property's cache target list
     * @param name The property name
     * @param pluginname The plugin's name
     * @return True if the push request was stopped
     */
    public boolean stopPushProperty(String name, String pluginname) {
        DataToken token = null;
        boolean removed = false;

        if ((token = (DataToken)propertytokens.get(name)) != null &&
            token.isOwner()) {
            PlugIn plugin = findPlugIn(pluginname);
            try {
                if (plugin != null) {
                    long sptime = System.currentTimeMillis();
                    plugin.uncacheProperty(name);
                    token.removeCacheTarget(pluginname);
                    removed = true;
                    if (logprops) System.err.println("DATA OVERHEAD REMOTE PUSH\t" + name + "\t" + pluginname + "\t" + (System.currentTimeMillis() - sptime));
               } else {
                    log.log("Could not find plugin " + pluginname + " to push " + name + " to");
                }
            } catch (RemoteException e) {
                log.log("Error pushing " + name + " to pluginname");
            }
        } else {
            log.log("Request push stop for " + name + " from " + pluginname + " denied because I am not the owner");
        }

        return removed;
    }

    /**
     * Pushes property data to the plugins designated as cache targets
     * for that property.
     * @param name The property name
     * @param value The value to push
     * @param token The property's data token object (optional)
     */
    public void pushProperty(String name, Object value, DataToken token) {

        if (token == null) {
            token = (DataToken)propertytokens.get(name);
        }
        if (token == null || !token.hasCacheTargets()) return;

        synchronized (token) {
            for (Iterator i = token.getCacheTargets(); i.hasNext(); ) {
                String pluginname = (String)i.next();
                PlugIn plugin = findPlugIn(pluginname);
                try {
                    if (plugin != null) {
                            long sptime = System.currentTimeMillis();
                        if (cachetoken == null) cachetoken = new DataToken(true, getPlugInName());
                        plugin.cacheProperty(name, value, cachetoken);
                        if (logprops) System.err.println("DATA OVERHEAD REMOTE PUSH\t" + name + "\t" + pluginname + "\t" + (System.currentTimeMillis() - sptime));
                    } else {
                        log.log("Error pushing " + name + " to " + pluginname + ", could not find plugin reference");
                    }
                } catch (RemoteException e) {
                    log.log("Error pushing " + name + " to " + pluginname + ", removing them as cache target");
                    i.remove();
                }
            }
        }
    }

    /**
     * If called, the "stop" flag is set for the plugin, so that it
     * will terminate operations at the earliest convenient point.
     */
    public void setStop() {

        done = true;
        setPulse(Simulator.PULSE_NONE);
        log.log("Starting to stop, please wait...", Log.LOG_INFO);
    }

    /**
     * Determines if the stop flag has been set or not
     */
    public boolean isStopping() {

        return done;
    }

    /**
     * If called, the "dead" flag is set for the plugin, declaring
     * the agent dead.
     */
    public void setDead() {

        dead = true;
    }

    /**
     * This will generally only be called by the simulator.
     * @see #setPulse(int, int)
     */
    public void setPulse(int type) {

        setPulse(type, 0);
    }

    /**
     * Notifies the plugin that it should perform a pulse of a particular
     * type (begin, end or in).  This will  generally only be called
     * by the simulator.  Internally, this sets a flag and notifies
     * the main thread that it can perform the pulse in question.
     * @param type The type of pulse to perform (Simulator.PULSE_BEGIN,
     * PULSE_END, PULSE_IN)
     * @param time The amount of time to allow processing in milliseconds
     * (PULSE_IN only)
     */
    public void setPulse(int type, int time) {

        // Are we done?
        if (isStopping())  {
            synchronized (pulselock) {
                pulselock.notify();
            }
            return;
        }

        // Notify the plugin thread
        synchronized (pulselock) {
            pulse = type;
            runtime = time;
            pulselock.notify();
        }
        
        // Wait for it to finish
        synchronized (pulsedone) {
            try {
                if (pulse != Simulator.PULSE_NONE) {
                    pulsedone.wait();
                }
            } catch (InterruptedException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        }
    }

    /**
     * The main control loop.  This simply loops until the agent's
     * stop flag has been set.
     */
    public void run() {

        // Control loop
        try {
            begin();
            while (!isStopping()) {
                try {
                    pulse();
                } catch (InterruptedException e) {
                    System.err.println("Exception: " + e.getMessage());
                    e.printStackTrace();
                }
            }
            end();
        } catch (RemoteException e) {
            System.err.println("RemoteException: " + e.getMessage());
            e.printStackTrace();

        } catch (Exception e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();

        } finally {
            if (!isStopping())
                setStop();
            setDead();
        }

        // Finish
        System.exit(0);
    }

    /**
     * Called when the plugin starts running.  This can be overriden
     * to allow for plugin-specific instantiation.
     * <p>
     * FarmPlugIn.begin() is where the simulator registration process
     * takes place, so if you do override this function, make sure
     * you add a call to super.begin() at it beginning, so that your
     * new component will actually connect to the simulator.
     * @see FarmSimulator#registerPlugIn
     */
    public void begin() throws RemoteException {
        
        // Register
        for (int i = 0; i < 10; i++) {
            try {
                simulator.registerPlugIn(this);
                break;
            } catch (ConnectException e) {
                System.err.println("Error connecting to simulator: " + e.getMessage());
                mysleep(2000);
            }
        }
    }

    /**
     * This is called periodically for pulse actions to
     * take place.  It takes pulse information, presumably provided
     * by the simulator, and uses it to route control the the
     * appropriate pulse* function.
     * @see #pulseBegin
     * @see #pulseIn
     * @see #pulseEnd
     */
    protected void pulse() throws InterruptedException, RemoteException {
        
	if (!isStopping()) {
	    synchronized (pulselock) {

		if (pulse == Simulator.PULSE_NONE) {
		    pulselock.wait();
		}

		switch (pulse) {
		case Simulator.PULSE_BEGIN:
		    pulseBegin();
		    break;

		case Simulator.PULSE_IN:
		    pulseIn(runtime);
		    break;

		case Simulator.PULSE_END:
		    pulseEnd();
                    cleanProperties();
		    break;
                
		case Simulator.PULSE_NONE:
		    break;
		}
	    }
	}

        synchronized (pulsedone) {
            pulse = Simulator.PULSE_NONE;
            pulsedone.notify();
        }
    }
    
    /**
     * This is called at the beginning of a pulse, before the
     * agents are started.
     */
    protected void pulseBegin() throws RemoteException {
    }

    /**
     * This may be called during the actual time of a pulse.  This
     * method represents the modeled runtime of the system, and
     * as such, the activities performed by the method should take
     * no longer than the time indicated.
     * @param time The amount of real time in ms which should be spent
     * performing activities in the function.
     */
    protected void pulseIn(int time) throws RemoteException {
    }

    /**
     * This is called at the end of a pulse, after the agents
     * should have completed.
     */
    protected void pulseEnd() throws RemoteException {
    }

    /**
     * This unregisters the plugin from the simulator.  You can also override
     * this to perform necessary termination functionality.
     * <p>
     * As with begin, this function communicates with the simulator to
     * unregister the plugin, so if you do override this function, include
     * a super call to this end() method to allow this to take place.
     * @see #begin
     */
    public void end() throws RemoteException {

        // Just in case
        synchronized (pulsedone) {
            pulse = Simulator.PULSE_NONE;
            pulsedone.notify();
        }

        // Unregister
        simulator.unregisterPlugIn(this);

        log.log("PlugIn Ending");
    }

    /**
     * Runs a function locally at the component, useful
     * for functions that require excessive bandwidth
     * @param function The function to perform locally
     * @return Returns whatever object is returned by
     *  the LocalFunction that is passed in.
     */
    public Object runLocalFunction(LocalFunction function) throws RemoteException {

        return function.run(this);
    }

    /**
     * Convenience function.  Returns the integer value
     * of the Simulator.P_TIME property.
     * @return The P_TIME value, or -1 if not found
     */
    public int getTime() {
        Integer i = (Integer)getProperty(Simulator.P_TIME);
        return (i != null) ? i.intValue() : -1;
    }

    /**
     * Convenience function.  Returns true if the time
     * has changed since the function was last called.
     * @return True if the time has changed, faled otherwise
     * @see #getTime
     */
    public boolean timeChanged() {
        int thistime = getTime();

        if (thistime != lasttime) {
            lasttime = thistime;
            return true;
        }

        return false;
    }

    /**
     * Convenience function.  Puts the plugin (thread) to sleep
     * for the specified period of time.
     * @param i The number of ms to sleep
     */
    public void mysleep(int i) {
        if (i < 0) return ;

        try {
            Thread.currentThread().sleep(i);
        } catch (InterruptedException e) { }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new FarmPlugIn(args);
            plugin.run();

       } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
