/**
 */

package farm.simulator;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.io.Serializable;

/**
 * This class represents a function that can be run "locally" 
 * at a property owner.  Frequently it is the case, especially
 * in data analysis routines, that a large number of discrete
 * state properties must be accessed by a particular function.
 * Because this data travels over RMI, the overhead can become
 * quite expenive.  This class provides an alternative, by allowing
 * the designer to send the code to the data, rather than bringing
 * the data to the code, which can provide a significant savings
 * in time and bandwith.
 * <p>
 * To use this functionality, create an object which implements
 * the LocalFunction interface.  Then, inside your plugin or agent,
 * pass the object to one or more components' runLocalFunction method.
 * This will perform the LocalFunction locally, and return whatever
 * is returned from the run function in the
 * interface.
 */
public interface LocalFunction extends Serializable {

    /**
     * Override this function to perform the activity
     * you want to execute locally at the component.
     * @param comp A property owner reference,
     *   used to access property information
     */
    public Object run(PropertyOwner owner) throws RemoteException;
}
